package com.digitalsanctuary.atg.servlet.pipeline;

import java.io.IOException;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import javax.servlet.ServletException;

import com.digitalsanctuary.atg.util.StringLengthReverseComparator;

import atg.core.util.StringUtils;
import atg.servlet.DynamoHttpServletRequest;
import atg.servlet.DynamoHttpServletResponse;
import atg.servlet.pipeline.InsertableServletImpl;

/**
 * Checks the request based on mime-type and URI, and determines the correct cache related response headers to return.
 * For instance you might want to cache images for a month, CSS files for a week, and JSPs not at all.
 * 
 * @author Devon Hillard devon@digitalsanctuary.com copyright 2008 DigitalSanctuary Inc. All Rights Reserved.
 * 
 */
public class CacheControllerPipelineServlet extends InsertableServletImpl {

    /**
     * This Map stores mime-types and their associated cache time (in seconds) from the configuration file.
     */
    private Map mMimeTypeCacheMap = new HashMap();

    /**
     * This Map stores partial or full URIs and their associated cache time (in seconds) from the configuration file.
     */
    private Map mURICacheMap = new Hashtable();

    /**
     * This Set stores an ordered collection of the partial or full URIs from the URICacheMap sorted longest to
     * shortest.
     */
    private Set mURICacheTreeSet = new TreeSet();

    /**
     * This Map holds a cache of cache time values associated to a specific URI. This means the cache time for a
     * specific asset only needs to be calculated using the mime-type and the URI map once.
     */
    private Map mCacheCache = new HashMap();

    /**
     * This setting determines if requests with query params are cachable. The default is to not cache any requests with
     * query params. If you set this to true, the cache time value will be the same for all requests to the base URI
     * (sans query params). What this means is that if you have a URI which returns different mime types based on query
     * params, you should leave this set to false.
     */
    private boolean mCacheRequestsWithQueryParams;

    /**
     * Checks the request based on mime-type and URI, and determines the correct cache related response headers to
     * return. For instance you might want to cache images for a month, CSS files for a week, and JSPs not at all.
     * 
     * @see atg.servlet.pipeline.PipelineableServletImpl#service(atg.servlet.DynamoHttpServletRequest,
     *      atg.servlet.DynamoHttpServletResponse)
     * @param pRequest
     *                the ATG Request.
     * @param pResponse
     *                the ATG Response.
     * @throws IOException
     *                 in case of error.
     * @throws ServletException
     *                 in case of error.
     */
    public void service(final DynamoHttpServletRequest pRequest, final DynamoHttpServletResponse pResponse)
	    throws IOException, ServletException {
	final String requestURI = pRequest.getRequestURI();
	// Check to ensure we have a URI to work with
	if (requestURI != null) {
	    if (isLoggingDebug()) {
		logDebug("CacheControllerPipelineServlet.service: entering with URI: " + requestURI);
	    }
	    // Default behavior is to not cache
	    Long cacheTime = new Long(0);
	    // Make sure that either the cacheRequestsWithQueryParams flag is set to true or that there are no query
	    // params.
	    if (isCacheRequestsWithQueryParams() || StringUtils.isBlank(pRequest.getQueryString())) {
		// Check the cache to see if this URI has already had it's cache time determined.
		if (mCacheCache.containsKey(requestURI)) {
		    // If so, use the cached time.
		    cacheTime = (Long) mCacheCache.get(requestURI);
		    if (isLoggingDebug()) {
			logDebug("CacheControllerPipelineServlet.service: cache time found in cache: " + cacheTime);
		    }
		} else {
		    // Start by working with the mime-type of the request.
		    final String mimeType = pRequest.getMimeType();
		    if (mimeType != null && mMimeTypeCacheMap.containsKey(mimeType)) {
			cacheTime = new Long((String) mMimeTypeCacheMap.get(mimeType));
		    }
		    // Then move on to checking the URI map for any entries which match the start of the current request
		    // URI.
		    final Iterator uriIterator = mURICacheTreeSet.iterator();
		    while (uriIterator.hasNext()) {
			final String currentURIEntry = (String) uriIterator.next();
			if (!StringUtils.isBlank(currentURIEntry) && requestURI.startsWith(currentURIEntry)) {
			    cacheTime = Long.valueOf((String) mURICacheMap.get(currentURIEntry));
			    break;
			}
		    }
		}
	    }
	    // Now that the cache time has been determined, set the appropriate response headers.
	    final String userAgent = pRequest.getHeader("User-Agent");
	    final boolean isIE = ((userAgent != null) && (userAgent.indexOf("MSIE") != -1));
	    final long now = System.currentTimeMillis();
	    pResponse.setDateHeader("Last-Modified", now);
	    pResponse.setHeader("ETag", Integer.toString((int) now));
	    pResponse.setDateHeader("Expires", now + cacheTime.longValue() * 1000);
	    pResponse.setHeader("Cache-Control", "max-age=" + cacheTime.longValue());
	    if (isIE) {
		pResponse.addHeader("Cache-Control", "pre-check=" + cacheTime.longValue());
		pResponse.addHeader("Cache-Control", "post-check=" + cacheTime.longValue());
	    }
	    // If the cache time is zero, do what we can to prevent the response from being cached.
	    if (cacheTime.longValue() == 0L) {
		pResponse.setHeader("Pragma", "no-cache");
		pResponse.setDateHeader("Expires", 0L);
		pResponse.setHeader("Cache-Control", "no-cache");
		pResponse.setHeader("Cache-Control", "no-store");
	    }
	    // Store the calculated cache time in the cache time cache.
	    if (isLoggingDebug()) {
		logDebug("CacheControllerPipelineServlet.service: adding cache time entry to the cache: URI: "
			+ requestURI + " and cache time: " + cacheTime);
	    }
	    mCacheCache.put(requestURI, cacheTime);
	}
	passRequest(pRequest, pResponse);
    }

    /**
     * This method flushes the URI-cache time cache.
     */
    public void flushCache() {
	mCacheCache = new HashMap();
    }

    /**
     * @return the mimeTypeCacheMap
     */
    public Map getMimeTypeCacheMap() {
	return this.mMimeTypeCacheMap;
    }

    /**
     * @param pMimeTypeCacheMap
     *                the mimeTypeCacheMap to set
     */
    public void setMimeTypeCacheMap(final Map pMimeTypeCacheMap) {
	this.mMimeTypeCacheMap = pMimeTypeCacheMap;
	flushCache();
    }

    /**
     * @return the uRICacheMap
     */
    public Map getURICacheMap() {
	return this.mURICacheMap;
    }

    /**
     * @param pCacheMap
     *                the uRICacheMap to set
     */
    public void setURICacheMap(final Map pCacheMap) {
	this.mURICacheMap = pCacheMap;
	// Sort the URI entries in reverse order of length, so the longest (and most specific) URIs are first in the
	// Set. This will optimize the checks in the service method and allow for a fast find-return.
	final TreeSet sortedURITreeSet = new TreeSet(new StringLengthReverseComparator());
	sortedURITreeSet.addAll(this.mURICacheMap.keySet());
	this.mURICacheTreeSet = sortedURITreeSet;
	flushCache();
    }

    /**
     * @return the cacheRequestsWithQueryParams
     */
    public boolean isCacheRequestsWithQueryParams() {
	return this.mCacheRequestsWithQueryParams;
    }

    /**
     * @param pCacheRequestsWithQueryParams
     *                the cacheRequestsWithQueryParams to set
     */
    public void setCacheRequestsWithQueryParams(final boolean pCacheRequestsWithQueryParams) {
	this.mCacheRequestsWithQueryParams = pCacheRequestsWithQueryParams;
    }
}
